module fault_analysis

 ! The routines for fault striation analysis
 !
 ! Programmer : A. Yamaji, K. Sato, and M. Otsubo
 !              (Division of Earth & Planetary Sciences, Kyoto Univ.)
 !
 ! Version    : 7.0  November, 2010  (A. Yamaji)

!/////////////////////////// Decraration Section ///////////////////////////////
use grid_points
use math_operators
use specify_file_names, only : get_path_filename_extension

implicit none
private

! public procedures
public MIM_main
  
! private 
type, private :: Tfault_data
  type (Tazpl)     :: dip        ! dip direction and dip
  type (Tazpl)     :: sst        ! striation
  character(len=1) :: sen        ! sense of movement (N, R, D, S)
  ! real (4)         :: st, dp, rk ! strike, dip & rake of focal mechanism
  real (4), dimension(3) :: p    ! normal to hanginwall (lower hemisphere)
  real (4), dimension(3) :: s    ! hangingwall's movement direction
end type Tfault_data

character(len=20), parameter :: MIM_main_version = '7.0'

character(len=300)  :: HomeDir
character(len=300)  :: DataFile
character(len=300)  :: ResultFile
integer (1)         :: kValue
integer (4)         :: Nrndm
logical (1)         :: RndSeed   ! if (.TRUE.) then call random_seed() else call random_seed(put=1)

integer (2)         :: NF = 0           ! number of fault
character(len=1)    :: DataType         ! 'g' = FDT, 's' = FMD
integer (8)         :: Niteration       ! number of iteration
type(Tfault_data), allocatable :: FT(:) ! instances of fault-slip data
integer (1), allocatable, dimension (:,:) :: fitness_table  ! (NGP,NF) -256 <=  <=255.
logical (1), allocatable, dimension (:,:) :: match_table    ! (NGP,NF)
integer (4)         :: Misfit_Threshold ! (degrees) read_pref_file assings a value
integer (2), dimension(0:180)  :: WFunc
integer (2), parameter         :: Max_WFunc = 200
character (len=40)             :: Func_Type
integer (4)                    :: Max_nCk

character(len=100)  :: Sender
character(len=200)  :: Message

integer (2), parameter :: IK1 = selected_int_kind(1)
integer (2), parameter :: IK2 = selected_int_kind(3)
integer (2), parameter :: IK4 = selected_int_kind(6)
integer (2), parameter :: SP  = selected_real_kind(p=6)
integer (2), parameter :: DP  = selected_real_kind(p=12)

real (4), parameter :: pi  = 3.1415926535897932384626433832795
real (4), parameter :: d2r = 0.017453292519943295769236907684886 
real (4), parameter :: r2d = 57.295779513082320876798154814105

!/////////////////////////////// Subprograms ///////////////////////////////////
contains

!-------------------------------------------------------------------------------
! Nrndm = 0 : entire nCk subsets are processed
! Nrndm > 0 : Nrndm subsets are rondomly chosen and processed
subroutine MIM_main( hd, df, rf, kv, gs, nr, rs )
character(len=300), intent(in) :: hd
character(len=300), intent(in) :: df
character(len=300), intent(in) :: rf
integer (1), intent(in)        :: kv
character(len=4), intent(in)   :: gs
integer (4), intent(in)        :: nr
logical (1), intent(in)        :: rs
character (len=12)  :: today1, thismoment1 ! job start time
character (len=12)  :: today2, thismoment2 ! job   end time
integer (1)         :: ierror
character (len=1)   :: yesorno

call date_and_time( date=today1, time=thismoment1 )
Homedir    = hd
DataFile   = df
ResultFile = rf 
kValue     = kv
Nrndm      = nr
RndSeed    = rs
if ((gs=='S').or.(gs=='s')) then
  DataType = 's'
  call read_FMD()
else
  DataType = 'g'
  call read_FDT()
end if

Sender  = "MIM_main"
call read_pref_file()

! get number of iteration
if (nr > 0) then
  Niteration = nr
else
  Niteration = binomial_coef(NF,kValue)
  if (Niteration < 0 .or. Niteration > Max_nCk) then
    Message = "Too large number of iteration"
    call application_terminate()
  end if
end if

call GP_read_CGP_file (ierror); if (ierror/=0) call application_terminate()
!call GP_read_NBS_file (ierror); if (ierror/=0) call application_terminate()
call determine_weightfunc ()
call allocate_fitness_and_match_tables ()
call fill_fitness_and_match_tables ()
call my_random_seed()

if (Nrndm == 0) then
  call MIM_all_combinations()
else
  if (DataType=='g') call MIM_random_g()
  if (DataType=='s') call MIM_random_s()
end if

call date_and_time( date=today2, time=thismoment2 )
call MIM_save_results( today1, thismoment1, &
                       today2, thismoment2 )
call finalize_allocatable_arrays()
call GP_finalize_allocatable_arrays()
end subroutine MIM_main

!-------------------------------------------------------------------------------
subroutine allocate_fitness_and_match_tables()
integer :: er

Sender = "allocate_fitness_and_match_tables"

if ( allocated(fitness_table) ) then
  fitness_table(:,:) = 0_IK2
else
  allocate ( fitness_table(1:NGP, 1:NF), stat = er )
  if (er /= 0)  then
    Message = "Memory allocation error [1]"
    call application_terminate()
  end if
end if

if ( allocated(match_table) ) then
  match_table(:,:) = .false.
else
  allocate ( match_table(1:NGP, 1:NF), stat = er )
  if (er /= 0) then
    Message = "Memory allocation error [2]"
    call application_terminate()
  end if
end if
end subroutine allocate_fitness_and_match_tables

!-------------------------------------------------------------------------------
subroutine application_terminate( line, unt )
integer  (4), optional, intent(in) :: line
integer  (1), optional, intent(in) :: unt
character(len=200)                 :: buf

call finalize_allocatable_arrays()
if ( present(unt) ) close(unit=unt)
print*
print*,"  Error in the subprogram "// trim(Sender)
if ( present(line) ) then
  write( buf, '(2x,a,i10)' ) trim(Message), line
else
  write( buf, '(2x,a    )' ) trim(Message)
end if
write(*,'(/,a)') trim(buf)
write(*,'(  a)') "  Execution terminated."
write(*,'(  a)') "  Press ENTER to close this window."
read*
stop
end subroutine application_terminate

!-------------------------------------------------------------------------------
subroutine determine_weightfunc ()
character(len=300), parameter   :: array_func_file = 'mim_array.txt'
character(len=3) :: line
real    (4)      :: dummy
integer (2)      :: i
integer          :: er

Sender = "determine_weightfunc"
open (unit = 1, file = trim(HomeDir)//"\"//trim(array_func_file), &
      status = 'old', action = 'read', iostat = er)
if (er /= 0) then
  Message = array_func_file // " not available."
  call application_terminate()
end if
read (1, *) (WFunc(i), i=0,180)
read (1, *) Func_Type
close(unit=1)

! test WFunc
do i = 1, 180
  if ((wfunc(i)<0) .or. (WFunc(i) > Max_WFunc)) then
    write(line,'(a3)') i
    Message = "Invalid number #" // line
    write(*,*) , wfunc(i), " at line ", i
    call application_terminate()
  end if
end do
end subroutine determine_weightfunc

!-------------------------------------------------------------------------------
! Purpose: converts fault normal and slip direction to FDT
subroutine faultnormalslipdirec2FDT( p, s, dd, ss, sn )
real (4), dimension(3), intent(in) :: p  ! fault normal
real (4), dimension(3), intent(in) :: s  ! slip direction
type(Tazpl), intent(out)           :: dd ! dip direction
type(Tazpl), intent(out)           :: ss ! slip orientation
character(len=1), intent(out)      :: sn ! sense of movement
real (4), dimension(3) :: v, u

! dip direction
if (p(3) < 0) then
  v =  p
else
  v = -p
end if
dd = azpl(v)
dd = azpl( mod(dd%az*r2d+180, 360.), 90.-dd%pl*r2d )

! slip orientation
if (s(3) < 0) then
  v =  s
else
  v = -s
end if
ss = azpl(v)
ss%az = ss%az * r2d
ss%pl = ss%pl * r2d

! sense of movement
if (s(3) > 0.) then
  sn = 'R'
elseif (s(3) < 0.) then
  sn = 'N'
else
  v = uvector(dd)
  u = cross_prod(v,s)
  if (u(3) < 0.) then
    sn = 'D'
  else
    sn = 'S'
  end if
end if
end subroutine faultnormalslipdirec2FDT

!-------------------------------------------------------------------------------
subroutine fill_fitness_and_match_tables ()
real (4), dimension (3) :: s_theoretical
real (4)                :: misfit_angle
integer (4)             :: id
integer (2)             :: iflt
real (4), parameter     :: deg = 57.2957795130
integer (2), parameter  :: fitness_min = -100
integer (2)             :: iaxis ! 1, 2, or 3

do iflt = 1_IK2, NF
  do id = 1_IK4, NGP
    s_theoretical = theoretical_slip_direction ( GP(id)%sg, FT(iflt)%p )
    misfit_angle  = angular_distance( s_theoretical, FT(iflt)%s ) * deg
    fitness_table(id, iflt) = wfunc( nint(misfit_angle) ) + fitness_min
    match_table(id, iflt) = (misfit_angle <= misfit_threshold)
  end do
end do
end subroutine fill_fitness_and_match_tables

!-------------------------------------------------------------------------------
subroutine finalize_allocatable_arrays()
if ( allocated(FT           ) ) deallocate (FT)
if ( allocated(fitness_table) ) deallocate (fitness_table)
if ( allocated(match_table  ) ) deallocate (match_table)  
end subroutine finalize_allocatable_arrays

!-------------------------------------------------------------------------------
function isSameSeisicEvent(s1,s2) result(res)
  integer (2), intent(in) :: s1, s2    ! (f1 <= f2) is assumed
  integer (2) :: f1, f2
  logical (1) :: res
  
  res = .FALSE.  
  
  if     (s1 < s2) then
    f1 = s1; f2 = s2
  elseif (s1 > s2) then
    f1 = s2; f2 = s2
  else
    return   !======> Exit subprogram
  end if
  
  if (f2-f1 >= 2) then ! f1 and f2 are not consecutive numbers.
    return 
  else                 ! f1 and f2 are consecutive numbers.
    if (mod(f1,2)==1) res = .TRUE.
  end if
end function isSameSeisicEvent
!-------------------------------------------------------------------------------
subroutine MIM4g ()
integer (2), allocatable, dimension(:,:) :: fitness_table2  ! (NGP,NF)
integer (2), allocatable, dimension(:)   :: fitnessArray    ! (NGP)
integer (2), allocatable, dimension(:)   :: sum2, sum3
integer (4), dimension ( 1 )             :: id_max_fitness
logical (1), allocatable, dimension(:)   :: homog3
integer (2) :: f1, f2, f3, f4
integer (4) :: id, id_max
integer (4) :: loopTotal
integer (4) :: loopCount

allocate ( fitness_table2(NGP,NF), fitnessArray(NGP), &
           sum2(NGP), sum3(NGP), homog3(NGP) )

fitness_table2 = fitness_table
call GP_clear_GPcount()
loopTotal = binomial_coef(NF, 4_IK1)
print*, "  *      Main Loop      *"

open(unit=6, carriagecontrol='FORTRAN')
loopCount = 0_IK4
do f1 =      1_IK2, NF - 3_IK2
do f2 = f1 + 1_IK2, NF - 2_IK2; sum2(:) = fitness_table2(:,f1) + &
                                          fitness_table2(:,f2)
do f3 = f2 + 1_IK2, NF - 1_IK2; sum3(:) = fitness_table2(:,f3) + sum2(:)
                                 homog3 = match_table(:,f1) .and. &
                                 match_table(:,f2) .and. &
                                 match_table(:,f3)
							     if ( .not. any(homog3) ) then
							       loopCount = loopCount + binomial_coef(NF-f3,1)
							       cycle; 
							     end if
do f4 = f3 + 1_IK2, NF
   loopCount = loopCount + 1_IK4
   fitnessArray(:) = sum3(:) + fitness_table2(:,f4)
   id_max_fitness  = maxloc( fitnessArray )
   id_max = id_max_fitness(1)
   if ( match_table(id_max,f4) .and. homog3(id_max) ) &
        GP(id_max)%count = GP(id_max)%count + 1_IK4
   if (mod(loopCount,10000) == 0) write (6,'("+",x,i10," /",i9)') loopCount, loopTotal
end do
end do
end do
end do
close(unit=6)

deallocate( fitness_table2, fitnessArray, sum2, sum3, homog3 )
end subroutine MIM4g

!-------------------------------------------------------------------------------
subroutine MIM4s()
  integer (2), allocatable, dimension(:,:) :: fitness_table2  ! (NGP,NF)
  integer (2), allocatable, dimension(:)   :: fitnessArray    ! (NGP)
  integer (2), allocatable, dimension(:)   :: sum2, sum3
  integer (4), dimension ( 1 )             :: id_max_fitness
  logical (1), allocatable, dimension(:)   :: homog3
  integer (2) :: f1, f2, f3, f4
  integer (4) :: id, id_max
  integer (4) :: loopTotal
  integer (4) :: loopCount

  allocate ( fitness_table2(NGP,NF), fitnessArray(NGP), &
             sum2(NGP), sum3(NGP), homog3(NGP) )

  fitness_table2 = fitness_table
  call GP_clear_GPcount()
  loopTotal = binomial_coef(NF, 4_IK1)
  print*, "  *      Main Loop      *"

  open(unit=6, CARRIAGECONTROL='FORTRAN')
  loopCount = 0_IK4
  do f1 =      1_IK2, NF - 3_IK2
  do f2 = f1 + 1_IK2, NF - 2_IK2; sum2(:) = fitness_table2(:,f1) + &
                                            fitness_table2(:,f2)
  do f3 = f2 + 1_IK2, NF - 1_IK2; sum3(:) = fitness_table2(:,f3) + sum2(:)
                                  homog3 = match_table(:,f1) .and. &
								           match_table(:,f2) .and. &
									       match_table(:,f3)
								  if ( .not. any(homog3) ) then; loopCount = loopCount + binomial_coef(NF-f3,1); cycle; end if
  do f4 = f3 + 1_IK2, NF
     loopCount = loopCount + 1_IK4
     if ( isSameSeisicEvent(f1,f2) ) cycle
     if ( isSameSeisicEvent(f2,f3) ) cycle
     if ( isSameSeisicEvent(f3,f4) ) cycle
     fitnessArray(:) = sum3(:) + fitness_table2(:,f4)
     id_max_fitness  = maxloc( fitnessArray )
	 id_max = id_max_fitness(1)
	 if ( match_table(id_max,f4) .and. homog3(id_max) ) &
          GP(id_max)%count = GP(id_max)%count + 1_IK4
     if (mod(loopCount,1000) == 0) then
	   write (6,'("+",x,i10," /",i9)') loopCount, loopTotal
	 end if
  end do
  end do
  end do
  end do
  close(unit=6)

  deallocate( fitness_table2, fitnessArray, sum2, sum3, homog3 )
end subroutine MIM4s

!-------------------------------------------------------------------------------
subroutine MIM5g ()
integer (2), allocatable, dimension(:,:) :: fitness_table2  ! (NGP,NF)
integer (2), allocatable, dimension(:)   :: fitnessArray    ! (NGP)
integer (2), allocatable, dimension(:)   :: sum2, sum3, sum4
integer (4), dimension ( 1 )             :: id_max_fitness
logical (1), allocatable, dimension(:)   :: homog3, homog4  ! (NGP)
integer (2) :: f1, f2, f3, f4, f5
integer (4) :: id, id_max
integer (4) :: loopTotal
integer (4) :: loopCount

allocate ( fitness_table2(NGP,NF), fitnessArray(NGP), &
           sum2(NGP), sum3(NGP), sum4(NGP), homog3(NGP), homog4(NGP) )

fitness_table2 = fitness_table
call GP_clear_GPcount()
loopTotal = binomial_coef(NF, 5_IK1)
print*, "  *      Main Loop      *"

open(unit=6, CARRIAGECONTROL='FORTRAN')
loopCount = 0_IK4
do f1 =      1_IK2, NF - 4_IK2
do f2 = f1 + 1_IK2, NF - 3_IK2; sum2(:) = fitness_table2(:,f1) + &
                                          fitness_table2(:,f2)
do f3 = f2 + 1_IK2, NF - 2_IK2; sum3(:) = fitness_table2(:,f3) + sum2(:)
                                homog3 = match_table(:,f1) .and. &
                                         match_table(:,f2) .and. &
                                         match_table(:,f3)
							    if ( .not. any(homog3) ) then
							      loopCount = loopCount + binomial_coef(NF-f3,2)
							      cycle
							    end if
do f4 = f3 + 1_IK2, NF - 1_IK2; sum4(:) = fitness_table2(:,f4) + sum3(:)
                                homog4 = match_table(:,f4) .and. homog3
                                if ( .not. any(homog4) ) then
                                  loopCount = loopCount + binomial_coef(NF-f4,1)
                                  cycle
                                end if
do f5 = f4 + 1_IK2, NF
  loopCount = loopCount + 1_IK4
  fitnessArray(:) = sum4(:) + fitness_table2(:,f5)
  id_max_fitness = maxloc( fitnessArray )
  id_max = id_max_fitness(1)
  if ( homog4(id_max) .and. match_table(id_max,f5) ) &
    GP(id_max)%count = GP(id_max)%count + 1_IK4
  if (mod(loopCount,10000) == 0) write (6,'("+",x,i10," /",i9)') loopCount, loopTotal
end do
end do
end do
end do
end do
close(unit=6)
deallocate( fitness_table2, fitnessArray, sum2, sum3, sum4, homog3, homog4 )
end subroutine MIM5g

!-------------------------------------------------------------------------------
subroutine MIM5s ()
  integer (2), allocatable, dimension(:,:) :: fitness_table2  ! (NGP,NF)
  integer (2), allocatable, dimension(:)   :: fitnessArray    ! (NGP)
  integer (2), allocatable, dimension(:)   :: sum2, sum3, sum4
  integer (4), dimension ( 1 )             :: id_max_fitness
  logical (1), allocatable, dimension(:)   :: homog3, homog4 ! (NGP)
  integer (2) :: f1, f2, f3, f4, f5
  integer (4) :: id, id_max
  integer (4) :: loopTotal
  integer (4) :: loopCount

  allocate ( fitness_table2(NGP,NF), fitnessArray(NGP), &
             sum2(NGP), sum3(NGP), sum4(NGP), homog3(NGP), homog4(NGP) )

  fitness_table2 = fitness_table
  call GP_clear_GPcount()
  loopTotal = binomial_coef(NF, 5_IK1)
  print*, "  *      Main Loop      *"

  open(unit=6, CARRIAGECONTROL='FORTRAN')
  loopCount = 0_IK4
  do f1 =      1_IK2, NF - 4_IK2
  do f2 = f1 + 1_IK2, NF - 3_IK2; sum2(:) = fitness_table2(:,f1) + &
                                            fitness_table2(:,f2)
  do f3 = f2 + 1_IK2, NF - 2_IK2; sum3(:) = fitness_table2(:,f3) + sum2(:)
                                  homog3 = match_table(:,f1) .and. &
								           match_table(:,f2) .and. &
										   match_table(:,f3)
								  if ( .not. any(homog3) ) then; loopCount = loopCount + binomial_coef(NF-f3,2); cycle; end if
  do f4 = f3 + 1_IK2, NF - 1_IK2; sum4(:) = fitness_table2(:,f4) + sum3(:)
                                  homog4 = match_table(:,f4) .and. homog3
								  if ( .not. any(homog4) ) then; loopCount = loopCount + binomial_coef(NF-f4,1); cycle; end if
  do f5 = f4 + 1_IK2, NF
     loopCount = loopCount + 1_IK4
     if ( isSameSeisicEvent(f1,f2) ) cycle
     if ( isSameSeisicEvent(f2,f3) ) cycle
     if ( isSameSeisicEvent(f3,f4) ) cycle
     if ( isSameSeisicEvent(f4,f5) ) cycle
     fitnessArray(:) = sum4(:) + fitness_table2(:,f5)
     id_max_fitness = maxloc( fitnessArray )
	 id_max = id_max_fitness(1)
	 if ( homog4(id_max) .and. match_table(id_max,f5) ) &
          GP(id_max)%count = GP(id_max)%count + 1_IK4
     !if (mod(loopCount,1000) == 0) write (6,'("+",i10," /",i10,i10," /",i5)') loopCount, loopTotal, f1, NF
	 if (mod(loopCount,1000) == 0) write (6,'("+",x,i10," /",i9)') loopCount, loopTotal
  end do
  end do
  end do
  end do
  end do
  close(unit=6)
  deallocate( fitness_table2, fitnessArray, sum2, sum3, sum4, homog3, homog4 )
end subroutine MIM5s

!-------------------------------------------------------------------------------
subroutine MIM_all_combinations ()
if ((kValue==4) .and. (DataType=='g')) call MIM4g()
if ((kValue==4) .and. (DataType=='s')) call MIM4s()
if ((kValue==5) .and. (DataType=='g')) call MIM5g()
if ((kValue==5) .and. (DataType=='s')) call MIM5s()
end subroutine MIM_all_combinations

!-------------------------------------------------------------------------------
subroutine MIM_random_g ()
integer (2), allocatable, dimension(:)   :: fitnessArray 
integer (2), allocatable, dimension(:,:) :: fitness_table2
integer (2), dimension (5) :: ifd
integer (4), dimension (1) :: id_max_fitness
integer (4) :: id_max
integer (4) :: iter = 0
integer (4) :: i,j
logical     :: homo

print*, "  *      Main Loop      *"
call GP_clear_GPcount()
allocate ( fitnessArray(1:NGP), fitness_table2(1:NGP,1:NF) )
do i = 1_IK4,NF
do j = 1_IK4,NGP
  fitness_table2(j,i) = fitness_table(j,i)
end do
end do
open(unit=6, carriagecontrol='FORTRAN')
do while (iter <= Nrndm)
  iter = iter + 1_IK4
  if (mod(iter,1000)==0) write (6, '("+",x,i10," /",i9)' ) iter, Nrndm
  call dice()
  fitnessArray(:) = fitness_table2(:,ifd(1))
  do i = 2_IK4, kValue
    fitnessArray(:) = fitnessArray(:) + fitness_table2(:,ifd(i))
  end do
  id_max_fitness = maxloc(fitnessArray)
  id_max = id_max_fitness(1)
  homo = .TRUE.
  do i = 1_IK4, kValue
    homo = homo .and. match_table(id_max,ifd(i))
  end do
  if (homo) GP(id_max)%count = GP(id_max)%count + 1_IK4
end do ! while
close(unit=6)
deallocate( fitnessArray, fitness_table2 )

contains !-----
subroutine dice() ! choose k different faults 
logical (1) :: flg(NF)
real    (4) :: r(3)  ! r has the precision of ~6 digits.
real    (4) :: s
integer (2) :: i,j
flg = .FALSE.
call random_number(r)
forall (i=1:3) flg(ceiling(NF*r(i))) = .TRUE.
do
  call random_number(s)
  flg(ceiling(NF*s)) = .TRUE.
  if (count(flg)==kValue) exit
end do
j = 1
do i = 1, NF
  if (flg(i)) then
    ifd(j) = i
    j = j + 1
  end if
end do
end subroutine dice
end subroutine MIM_random_g

!-------------------------------------------------------------------------------
subroutine MIM_random_s ()
integer (2), allocatable, dimension(:)   :: fitnessArray 
integer (2), allocatable, dimension(:,:) :: fitness_table2
integer (2), dimension (5) :: ifd
integer (4), dimension (1) :: id_max_fitness
integer (4) :: id_max
integer (4) :: iter = 0
integer (4) :: i,j
logical     :: homo

print*, "  *      Main Loop      *"
call GP_clear_GPcount()                                
allocate ( fitnessArray(1:NGP), fitness_table2(1:NGP,1:NF) )
do i = 1_IK4,NF
do j = 1_IK4,NGP
  fitness_table2(j,i) = fitness_table(j,i)
end do
end do
open(unit=6, carriagecontrol='FORTRAN') 
do while (iter <= Nrndm)
  iter = iter + 1_IK4
  if (mod(iter,1000)==0) write (6, '("+",x,i10," /",i9)' ) iter, Nrndm
  call dice()
  !write(*,*) iter, (ifd(j),j=1,kValue)
  fitnessArray(:) = fitness_table2(:,ifd(1))
  do i = 2, kValue
    fitnessArray(:) = fitnessArray(:) + fitness_table2(:,ifd(i))
  end do
  id_max_fitness = maxloc(fitnessArray)
  id_max = id_max_fitness(1)
  homo = .TRUE.
  do i = 1_IK4, kValue
    homo = homo .and. match_table(id_max,ifd(i))
  end do
  if (homo) GP(id_max)%count = GP(id_max)%count + 1_IK4
end do ! while
close(unit=6)
deallocate( fitnessArray, fitness_table2 )

contains !-----
subroutine dice() ! choose k different faults 
logical (1) :: flg(NF)
integer (2) :: nfh   ! NF / 2
real    (4) :: r(3)  ! r has the precision of ~6 digits.
real        :: s
integer (2) :: ifdh(5)
integer (2) :: i,j
flg = .FALSE.
nfh = NF / 2
call random_number(r)
forall (i=1:3) flg(ceiling(nfh*r(i))) = .TRUE.
do
  call random_number(s)
  flg(ceiling(nfh*s)) = .TRUE.
  if (count(flg)==kValue) exit
end do
j = 1
do i = 1, nfh
  if (flg(i)) then
    ifdh(j) = i
    j = j + 1
  end if
end do
do i = 1, kValue
  call random_number(s)
  ifd(i) = 2 * ifdh(i) - nint(s)
end do
end subroutine dice
end subroutine MIM_random_s

!-------------------------------------------------------------------------------
subroutine MIM_save_results( job_start_day, job_start_time, &
                             job_end_day,   job_end_time )
  character (len=12), intent(in) :: job_start_day, job_start_time
  character (len=12), intent(in) :: job_end_day,   job_end_time
  character(len=1000):: cgpfile    
  character(len=256) :: name_cgp, ext_cgp
  real    (4)  :: dummy
  integer (4)  :: id        ! ID number of gridpoint
  integer (2)  :: iflt      ! ID number of fault
  type (Tazpl) :: ax1, ax3  ! azimuth and plunge of sigma1 and sigma3 axes
  integer      :: er

  Sender = "MIM_save_results"
  open (unit = 2, file = ResultFile, action = 'write', iostat = er)
  if (er /= 0) then
    Message = "File I/O error"
  end if

  write (2, *) NF
  do iflt = 1_IK2, NF
    write (2, '(2f5.1,f6.1,f5.1,a2)') &
    FT(iflt)%dip%az, FT(iflt)%dip%pl, &
    FT(iflt)%sst%az, FT(iflt)%sst%pl, FT(iflt)%sen
  end do

  write (2,*) NGP
  write (2,*) count (GP(:)%count /= 0)

  cgpfile = CGP_file
  call get_path_filename_extension(cgpfile, name_cgp, ext_cgp)
  write (2, '(a)') trim(name_cgp) // trim(ext_cgp)
   
  call write_all_GPcount(2_IK1)

  write (2, '(a)') "Data = " // trim(DataFile)
  write (2, '(" k = ",i2)') kValue
  write (2, '(" Nrandom = ",i)') Nrndm
  if (RndSeed) then
    write(2,*) "Random_Seed()"
  else
    write(2,*) "Random_Seed(1)"
  end if
  write (2, '("Function = ",a)') trim(Func_Type)

  write (2, '("MI main ver. ",a)') trim(MIM_main_version)
  write (2, '("misft_threshold     = ", I)') Misfit_Threshold
 
  write (2, '("Job Start = ",a4,"/",a2,"/",a2,x,a2,":",a2,":",a6)') &
                                job_start_day(1:4),  &
                                job_start_day(5:6),  &
								job_start_day(7:8),  &
                                job_start_time(1:2), &
								job_start_time(3:4), &
								job_start_time(5:10)
  write (2, '("Job Eend = ",a4,"/",a2,"/",a2,x,a2,":",a2,":",a6)') &
                                job_end_day(1:4),    &
                                job_end_day(5:6),    &
								job_end_day(7:8),    &
                                job_end_time(1:2),   &
								job_end_time(3:4),   &
								job_end_time(5:10)
  close(unit = 2)
end subroutine MIM_save_results

!-------------------------------------------------------------------------------
subroutine my_random_seed()
integer              :: n,i
integer, allocatable :: sd(:)
if (RndSeed) then 
  call random_seed()
else
  call random_seed(size=n)
  allocate (sd(n))
  sd = 1
  call random_seed(put=sd)
  deallocate(sd)
end if
end subroutine my_random_seed

!-------------------------------------------------------------------------------
subroutine read_FDT()
integer   (4)     :: iflt
real      (4)     :: faz, fpl, saz, spl
character (len=1) :: sen
integer           :: ioer
integer   (4)     :: line
integer   (1)     :: unt
type(Tazpl)       :: dip, striae

Sender = "read_FDT"
open ( unit=1, file=DataFile, status='old', action='read', iostat=ioer )
if(ioer/= 0) then
  Message = "Data file not found"
  call application_terminate()
end if

! counts the number of fault and checks data format
NF = 0
do while (.not. eof(1))
  read (1, *, iostat=ioer) faz, fpl, saz, spl, sen
  if (ioer /= 0) then
    Message = "Invalid data at line number"
    call application_terminate(line = NF+1, unt = 1)
  elseif (scan("NRSDnrsd", sen) == 0) then
    Message = "Invalid data at line number"
    call application_terminate(line = NF+1, unt = 1)
  end if
  NF = NF + 1
end do

allocate( FT(1:NF), stat=ioer )
if (ioer /= 0) then
  Message = "Memory allocation error"
  call application_terminate()
end if

rewind(unit=1)
do iflt = 1, NF
  read (1, *) FT(iflt)%dip%az, FT(iflt)%dip%pl, FT(iflt)%sst%az, FT(iflt)%sst%pl, FT(iflt)%sen
  dip        = azpl( FT(iflt)%dip%az*d2r, FT(iflt)%dip%pl*d2r )
  striae     = azpl( FT(iflt)%sst%az*d2r, FT(iflt)%sst%pl*d2r )
  FT(iflt)%p = uvector( dip_azpl_to_pole_azpl( dip ) )
  FT(iflt)%s = uvector( striae )
  select case (FT(iflt)%sen) ! invert slip vectors in order to represent
  ! case ('N', 'n')          ! the vector of the hangingwalls
    case ('R', 'r')
      FT(iflt)%s = - FT(iflt)%s
    case ('D', 'd')
      if ( first_az_is_left_of_second_az(dip%az, striae%az) < 0_SP ) &
        FT(iflt)%s = - FT(iflt)%s
    case ('S', 's') 
       if ( first_az_is_left_of_second_az(dip%az, striae%az) > 0_SP ) &
        FT(iflt)%s = - FT(iflt)%s
  end select
end do

close(unit=1)
end subroutine read_FDT

!-------------------------------------------------------------------------------
subroutine read_FMD()
integer   (4)     :: idat
real      (4)     :: stk, dip, rak
integer           :: ioer
integer   (4)     :: line
integer   (1)     :: unt
real (4), dimension (3,3):: RotMat
real (4), dimension (3), parameter :: v = (/1.,0.,0./)
type(Tazpl) :: fap, sap

Sender = "read_FMD"
open ( unit=1, file=DataFile, status='old', action='read', iostat=ioer )
if (ioer /= 0) then
  Message ="Data file not found"
  call application_terminate()
end if

! counts the number of fault and checks data format
NF = 0
do while (.not. eof(1))
  read (1, *, iostat=ioer) stk, dip, rak
  if (ioer /= 0) then
    Message = "Invalid data at line number"
    call application_terminate(line = NF+1, unt = 1)
  end if
  NF = NF + 1
end do

allocate( FT(1:2*NF), stat=ioer )
if (ioer /= 0) then
  Message = "Memory allocation error"
  call application_terminate()
end if

rewind(unit=1)
do idat = 1, NF
  read(1,*) stk, dip, rak
  FT(2*idat-1)%p = uvector( azpl( (stk-90.)*d2r, (90.-dip)*d2r ) )
  RotMat = orthogonal_tensor_R( dip*d2r, (90.-stk)*d2r, rak*d2r )
  FT(2*idat-1)%s = matmul(RotMat, v)
  FT(2*idat)%p = FT(2*idat-1)%s
  FT(2*idat)%s = FT(2*idat-1)%p
  if (FT(2*idat)%p(3) > 0) then
    FT(2*idat)%p = -FT(2*idat)%p
    FT(2*idat)%s = -FT(2*idat)%s
  end if
  call faultnormalslipdirec2FDT( FT(2*idat-1)%p,   FT(2*idat-1)%s, &
                                 FT(2*idat-1)%dip, FT(2*idat-1)%sst, FT(2*idat-1)%sen )
  call faultnormalslipdirec2FDT( FT(2*idat  )%p,   FT(2*idat  )%s, &
                                 FT(2*idat  )%dip, FT(2*idat  )%sst, FT(2*idat  )%sen )
end do
NF = 2 * NF

close( unit=1 )
end subroutine read_FMD

!-------------------------------------------------------------------------------
subroutine read_pref_file ()
character(len=300) :: filename
character(len=300) :: preffile
integer (4)        :: threshold
integer            :: ioerror

Sender = "read_pref_file"

preffile = trim(HomeDir) // "\pref.txt"
open(unit=1, file=preffile, status='old', action='read', iostat=ioerror)
if (ioerror /= 0) then 
  Message = "File 'pref.txt' not found"
  call application_terminate()
end if
  
read(unit=1, fmt='(a)', iostat=ioerror) filename ! read CGP file name
if (ioerror /= 0) then
  Message = "Invalid pref file [1]"
  call application_terminate(unt=1)
end if
filename = trim(HomeDir) // "\" // trim(filename)
call GP_set_CGP_filename (filename)

read(unit=1, fmt='(i)', iostat=ioerror) threshold ! read misfit threshold
if (ioerror /= 0) then
  Message = "Invalid pref file [2]"
  call application_terminate(unt=1)
end if
Misfit_Threshold = threshold
  
read(unit=1, fmt='(i)', iostat=ioerror) Max_nCk
if (ioerror /= 0) then
  Message = "Invalid pref file [3]"
  call application_terminate(unt=1)
end if

read(unit=1, fmt='(a)', iostat=ioerror) filename ! read NBS file name
if (ioerror /= 0) then
  Message = "Invalid pref file [4]"
  call application_terminate(unt=1)
end if
filename = trim(HomeDir) // "\" // trim(filename)
call GP_set_NBS_filename (filename)

close( unit=1 )
end subroutine read_pref_file

!-------------------------------------------------------------------------------
function theoretical_slip_direction(sg, n) result (s)
real (4), dimension (3,3) :: sg ! stress tensor (sigma)
real (4), dimension   (3) :: n  ! unit vector normal to fault plane
real (4), dimension   (3) :: s  ! unit vector parallel to slip direction

s = projection_onto_plane ( n, matmul( sg, n ) )
s = normalize_vector (s)
end function theoretical_slip_direction

!-------------------------------------------------------------------------------
! File must have been opened with unit=unt before this subroutine is called.
subroutine write_all_GPcount(unt)
  integer (1), intent(in) :: unt
  integer (4)             :: id

  do id = 1_IK4, NGP
    if (GP(id)%count /= 0) write (unit=unt, fmt='(2i9)') id, GP(id)%count
  end do
end subroutine write_all_GPcount

!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
end module fault_analysis