module grid_points

!  Purpose:
!    To provide procedures using the computational grid points for 
!    fault-slip analysis. 
!
!  Record of revisions:
!    Date          Programmer         Description of change
!    ==========    ==============     =====================
!    11/28 2010    A. Yamaji          Original code (most of subprograms were modified from 
!                                     fault_analysis.f90 (ver. 6)

use math_operators
implicit none
private

! public subprograms
public GP_clear_GPcount
public GP_finalize_allocatable_arrays
public GP_read_CGP_file
public GP_read_NBS_file
public GP_set_CGP_filename
public GP_set_NBS_filename

character(len=300), public :: CGP_file        ! grid points
character(len=300), public :: NBS_file        ! neighbors

integer (4), public :: NGP                    ! number of grid points
type :: Tgridpoint
  real    (4), dimension (3,3) :: sg          ! stress tensor
  integer (4)                  :: count       ! fitness or number
  real    (4), dimension (3,3) :: princorient ! principal orientations (column vectors)
end type Tgridpoint
type(Tgridpoint), allocatable, public  :: GP(:)

type :: Tneighbors  ! stored in a linked-list
  integer (4) :: id
  type(Tneighbors), pointer :: next
end type Tneighbors
Type(Tneighbors), dimension(:), pointer :: NBS 
! NBS(i) is the head of the linked-list that stores the neighbors of the i-th point
! The following piece of code shows the method to access the neighbors of the 99th grid point. 
!
! type(Tneighbors), pointer :: ptr
! ptr => NBS(99)
! do
!   if(.not.associated(ptr)) exit
!   print*, ptr%id
!   ptr => ptr%next
! end do

integer (2), parameter :: IK1 = selected_int_kind(1)
integer (2), parameter :: IK2 = selected_int_kind(3)
integer (2), parameter :: IK4 = selected_int_kind(6)
integer (2), parameter :: SP  = selected_real_kind(p=6)
integer (2), parameter :: DP  = selected_real_kind(p=12)

contains !/////////////////////////////////////////////////////////////////////

!-------------------------------------------------------------------------------
subroutine GP_clear_GPcount ()
GP(:)%count = 0
end subroutine GP_clear_GPcount

!-------------------------------------------------------------------------------
subroutine GP_finalize_allocatable_arrays()
! if ( allocated(GP ) ) deallocate (GP)
! if ( allocated(NBS) ) deallocate (NBS)
deallocate (GP)
!deallocate (NBS)
end subroutine GP_finalize_allocatable_arrays

!-------------------------------------------------------------------------------
! Purpose: read NBS_file and stores the ID numbers of neighbors in the array of 
!          linked-lists, NBS()

subroutine GP_read_NBS_file (ierror)
integer (1),        intent(out) :: ierror
type(Tneighbors),   pointer     :: tail, ptr
integer (4) :: itemp
integer (4) :: line = 0
integer (4) :: nb, id ! neibhgors of the id-th grid point
integer (2) :: i
integer     :: ier

open(unit=9, file=NBS_file, status='OLD', action='READ', iostat=ier)
if (ier/=0) then
  print*, "Execution terminated in GP_read_NBS_file [1]"
  close(unit=9)
  ierror = 1
  return
end if

read(9,*) NGP;     line = line + 1
allocate(NBS(1:NGP), stat=ier)
if (ier/=0) then
  print*, "Execution terminated in GP_read_NBS_file [2]"
  print*, "Memory error---#",line
  close(unit=9)
  ierror = 1
  return
end if

open(unit=6, carriagecontrol='FORTRAN')
do id = 1, NGP
  read(9,*) nb;    line = line + 1
  read(9,*) itemp; line = line + 1 
  tail => NBS(id)
  tail%id = itemp
  nullify(tail%next)
  do i = 2, nb
    read(9,*) itemp; line = line + 1 
    allocate(tail%next, stat=ier)
    if (ier/=0) then
      print*, "Execution terminated in GP_read_NBS_file [3]"
      print*, "Memory error---#",line
      close(unit=9)
      ierror = 1
      return
    end if
    tail => tail%next
    nullify(tail%next)
    tail%id = itemp
  end do
  if (mod(id,3000)==0) & 
     write (6,'("+","   Generating computational grid points  ",x,i3,"% completed")') &
     nint( 100.0*real(id)/NGP )
end do ! id
close(unit=6)

close(unit=9)
ierror = 0 ! Executed successfully.
end subroutine GP_read_NBS_file

!-------------------------------------------------------------------------------
subroutine finalize_grid_points()
deallocate(NBS)
end subroutine finalize_grid_points

!-------------------------------------------------------------------------------
subroutine GP_set_CGP_filename(fn)
character(len=300), intent(in) :: fn
CGP_file = fn
end subroutine GP_set_CGP_filename

!-------------------------------------------------------------------------------
subroutine GP_set_NBS_filename(fn)
character(len=300), intent(in) :: fn
NBS_file = fn
end subroutine GP_set_NBS_filename

!-------------------------------------------------------------------------------
subroutine GP_read_CGP_file (er)
integer (1), intent(out) :: er
integer (4) :: igp
type(Tazpl) :: ap1, ap3
real (4)    :: e1(3), e2(3), e3(3)
real (4)    :: Q(3,3), sg0(3,3)
real (4)    :: sr   ! stress ratio Phi
character(len=5)    :: line
integer             :: ioerror
character(len=100)  :: msg
real (4), parameter :: d2r = 0.017453292  ! = pi/180

open(unit=1, file=cgp_file, status='old', action='read', iostat=ioerror)
if (ioerror /= 0) then
  msg = "Error---GP_read_CGP_file [1]"
  call showmessage()
  return
end if

read (1, *, iostat=ioerror) NGP
if (ioerror /= 0) then
  msg = "Error---GP_read_CGP_file [2] Line #1"
  call showmessage()
  return
end if

allocate( GP(1:NGP), stat=ioerror )
if (ioerror /= 0) then
  msg = "Error---GP_read_CGP_file [3]  Memory allocation error"
  call showmessage()
  return
end if

sg0(:,:) = 0_SP
sg0(1,1) = 1_SP

do igp = 1, NGP
  read (1, *, iostat=ioerror) ap1%az, ap1%pl, ap3%az, ap3%pl, sr
  if (ioerror /= 0) then
    write(line,'(a5)') igp+1
    msg = "Error---GP_read_CGP_file [3] Line #" // line
    call showmessage()
    return
  end if
  ap1%az = ap1%az * d2r
  ap1%pl = ap1%pl * d2r
  ap3%az = ap3%az * d2r
  ap3%pl = ap3%pl * d2r
  e1 = uvector( ap1 )
  e3 = uvector( ap3 )
  e2 = cross_prod ( e3, e1 )
  GP(igp)%princorient(:,1) = e1
  GP(igp)%princorient(:,2) = e2
  GP(igp)%princorient(:,3) = e3
  Q(1,1:3) = (/ e1(1), e2(1), e3(1) /)
  Q(2,1:3) = (/ e1(2), e2(2), e3(2) /)
  Q(3,1:3) = (/ e1(3), e2(3), e3(3) /)
  sg0(2,2) = sr
  GP(igp)%sg = matmul( matmul(Q, sg0), transpose(Q) )
end do
close(unit=1)
er = 0

contains !-------------------
subroutine showmessage()
print*, msg
close(unit=1)
er = 1
end subroutine showmessage

end subroutine GP_read_CGP_file

!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------

end module grid_points
