module math_operators

 ! Purpose    : Defines subprograms for derectional operations.
 !              The 1st, 2nd, and 3rd coordinate axes are oriented in the
 !              east, north, and upward, respectively.
 !
 ! Programmer : A. Yamaji and K. Sato
 !              (Division of Earth & Planetary Sciences, Kyoto Univ.)

!/////////////////////////// Decraration Section ///////////////////////////////

  implicit none
  private

! public data =================

  type, public :: Tazpl        ! azimuth and plunge
    real (4) :: az
    real (4) :: pl 
  end type Tazpl

  type, public :: Tspherical   ! spherical coordinates
    real (4) :: co  ! colatitude
    real (4) :: lo  ! longitude
  end type Tspherical

  real (4), parameter, dimension  (3)  :: Zero_Vector     = (/ 0, 0, 0 /)
  real (4), parameter, dimension (3,3) :: Identity_Tensor = reshape(       &
                                                            (/ 1, 0, 0,    &
                                                               0, 1, 0,    &
                                                               0, 0, 1 /), &
															(/3,3/) )
! public subprograms ==========

  interface azpl
    module procedure a_p_to_azpl
    module procedure spherical_to_azpl
    module procedure uvector_to_azpl
  end interface

  interface uvector    ! unit vector
    module procedure azpl_to_uvector
    module procedure spherical_to_uvector
  end interface

  interface spherical
    module procedure azpl_to_spherical
    module procedure uvector_to_spherical
  end interface
  
  public angular_distance
  public azpl
  public binomial_coef
  public cross_prod
  public dip_azpl_to_pole_azpl
  public first_az_is_left_of_second_az
  public get_a_random_number
  public normalize_vector
  public opposite_azimuth
  public orthogonal_tensor_Q
  public orthogonal_tensor_R
  public projection_onto_plane
  public spherical
  public uvector

! private data ================

  real (4), parameter, private :: pi      = 3.14159265359
  real (4), parameter, private :: half_pi = 1.57079632679 ! = pi / 2
  real (4), parameter, private :: deg = 0.01745329251994  ! = 180 / pi
  real (4), parameter, private :: eps = epsilon(deg)
  integer (2), parameter, private :: SP = selected_real_kind(p=6)
  integer (2), parameter, private :: DP = selected_real_kind(p=12)

contains !///////////////////// Subprograms ////////////////////////////////////

!-------------------------------------------------------------------------------

function angular_distance (a, b) result (angle) ! radians
  real (4), dimension (3), intent(in) :: a, b
  real (4)                            :: angle
  real (4) :: cosangle

  cosangle = dot_product( normalize_vector(a), normalize_vector(b) )
  if      (cosangle > 1.0_SP) then 
    cosangle =  1.0_SP
  else if (cosangle <-1.0_SP) then
    cosangle = -1.0_SP
  end if

  angle = acos(cosangle)
end function angular_distance

!-------------------------------------------------------------------------------
function a_p_to_azpl (a, p) result (x)
  real (4), intent(in) :: a, p
  type (Tazpl)         :: x
 
  x%az = a
  x%pl = p
end function a_p_to_azpl

!-------------------------------------------------------------------------------

function azpl_to_spherical (x) result(s)
  type (Tazpl), intent(in) :: x
  type (Tspherical)        :: s

  s%lo = half_pi - x%az
  s%co = half_pi + x%pl
end function azpl_to_spherical

!-------------------------------------------------------------------------------

function azpl_to_uvector (x) result (v)
  type (Tazpl), intent(in) :: x
  real (4), dimension (3)  :: v

  v = spherical_to_uvector( azpl_to_spherical(x) )
end function azpl_to_uvector

!-------------------------------------------------------------------------------
  ! Purpose : Calculates binomial coefficient (n, k).
  !           If the result is greater than Huge(integer(4)) then returns -1.

function binomial_coef ( n, k ) result (res)
  integer (2), intent(in) :: n
  integer (1), intent(in) :: k
  integer (8)             :: res
  real (8)    :: bound
  real (8)    :: w
  real (8)    :: v
  integer (1) :: i

  bound = dlog( real(huge(res), DP) )
  w = 0.0_DP
  v = 0.0_DP

  do i = 0, k - 1
    w = w + dlog( real(n-i, DP) )
    v = v + dlog( real(i+1, DP) )
  end do

  if ((w-v) < bound) then
    res = nint( dexp( w - v ) )  
  else
    res = -1
  end if
end function binomial_coef

!-------------------------------------------------------------------------------

function cross_prod (a, b) result (c)
  real (4), dimension (3), intent(in) :: a, b
  real (4), dimension (3)             :: c

  c(1) = a(2)*b(3) - a(3)*b(2)
  c(2) = a(3)*b(1) - a(1)*b(3)
  c(3) = a(1)*b(2) - a(2)*b(1)
end function cross_prod

!-------------------------------------------------------------------------------

function dip_azpl_to_pole_azpl (dip) result (pole)
  type (Tazpl), intent(in) :: dip
  type (Tazpl)             :: pole

  pole%az = pi + dip%az
  pole%pl = half_pi - dip%pl
end function dip_azpl_to_pole_azpl

!-------------------------------------------------------------------------------
  ! Purpose : Returns the angular distance between the given azimuths.
  !           If the first one is on the left side of the second, returns
  !           positive angle in radians. 

function first_az_is_left_of_second_az (az1, az2) result (angle)
  real (4), intent(in)     :: az1, az2
  real (4)                 :: angle
  real (4), dimension (3)  :: vector1, vector2
  real (4), dimension (3)  :: upward

  vector1 = uvector( azpl(az1, 0.0) )
  vector2 = uvector( azpl(az2, 0.0) )
  upward  = cross_prod( vector2, vector1 )
  angle   = asin( upward(3) )
end function first_az_is_left_of_second_az

!-------------------------------------------------------------------------------
  ! Purpose : Generates a sequence of pseudo random numbers in the range [0, 1) 
  !           congruential  method. It is important for this function that it 
  !           by generates the same sequence every time.
 
function get_a_random_number () result (res)
  real (4)            :: res
  real (4), save      :: x
  real (4), parameter :: m1 = 19201237
  real (4), parameter :: m2 =  1174097

  x   = mod( x + m1, m2 )
  res = x / m2
end function get_a_random_number

!-------------------------------------------------------------------------------

function normalize_vector(x) result(y)
  real (4), dimension (3), intent(in) :: x
  real (4), dimension (3)             :: y
  real (4)                            :: r
  real (4), parameter :: eps2 = 2_SP * eps

  r = sqrt( dot_product( x, x ) )
  if (r < eps2) then
    y = 0.0_SP
  else
    y = x / r
  end if
end function normalize_vector

!-------------------------------------------------------------------------------

function opposite_azimuth (a) result(res)
  type (Tazpl), intent(in) :: a
  type (Tazpl)             :: res
  real (4), parameter :: pi2 = pi * 2_SP
  real (4), parameter :: pi4 = pi * 4_SP

  res%az =   a%az + pi
  res%pl = - a%pl
  res%az = mod( res%az + pi4, pi2)
end function opposite_azimuth

!-------------------------------------------------------------------------------

function orthogonal_tensor_Q (theta, phi, psi) result (Q)
  real (4), intent(in)     :: theta, phi, psi
  real (4), dimension (3,3):: Q
  real (4) :: cos_tht, cos_phi, cos_psi
  real (4) :: sin_tht, sin_phi, sin_psi

  cos_tht = cos(theta); sin_tht = sin(theta)
  cos_phi = cos(phi);   sin_phi = sin(phi)
  cos_psi = cos(psi);   sin_psi = sin(psi)

  Q(1,1) =  cos_psi * cos_phi - sin_psi * cos_tht * sin_phi;
  Q(1,2) =  cos_psi * sin_phi + sin_psi * cos_tht * cos_phi;
  Q(1,3) =                      sin_psi * sin_tht;

  Q(2,1) = -sin_psi * cos_phi - cos_psi * cos_tht * sin_phi;
  Q(2,2) = -sin_psi * sin_phi + cos_psi * cos_tht * cos_phi;
  Q(2,3) =                      cos_psi * sin_tht;

  Q(3,1) =                                sin_tht * sin_phi;
  Q(3,2) =                              - sin_tht * cos_phi;
  Q(3,3) =                                cos_tht;
end function orthogonal_tensor_Q

!-------------------------------------------------------------------------------

function orthogonal_tensor_R (theta, phi, psi) result (R)
  real (4), intent(in)     :: theta, phi, psi
  real (4), dimension (3,3):: R

  R = transpose( orthogonal_tensor_Q(theta, phi, psi) )
end function orthogonal_tensor_R

!-------------------------------------------------------------------------------

function projection_onto_plane (n, x) result (y)
  real (4), dimension (3), intent(in) :: n ! unit vector normal to plane
  real (4), dimension (3), intent(in) :: x ! vector to project
  real (4), dimension (3)             :: y ! result of projection
  real (4), dimension (3,3) :: projector

  projector = Identity_Tensor - spread(n, 1, 3) * spread(n, 2, 3)
  y = matmul( projector, x )
end function projection_onto_plane

!-------------------------------------------------------------------------------

function spherical_to_azpl (s) result(x)
  type (Tspherical), intent(in) :: s
  type (Tazpl)                  :: x

  x%az = half_pi - s%lo
  x%pl = s%co - half_pi
end function spherical_to_azpl

!-------------------------------------------------------------------------------

function spherical_to_uvector (s) result (v)
  type (Tspherical), intent(in) :: s
  real (4), dimension (3)       :: v

  v(1) = sin(s%co) * cos(s%lo)
  v(2) = sin(s%co) * sin(s%lo)
  v(3) = cos(s%co)
end function spherical_to_uvector

!-------------------------------------------------------------------------------

function uvector_to_azpl (v) result(x)
  real (4), dimension (3), intent(in) :: v
  type (Tazpl)                        :: x
  real (4), parameter :: pi2 = pi * 2_SP
  real (4), parameter :: pi4 = pi * 4_SP
  
  x = spherical_to_azpl( uvector_to_spherical(v) ) 
  x%az = mod( x%az + pi4, pi2 )
end function uvector_to_azpl

!-------------------------------------------------------------------------------

function uvector_to_spherical (v) result (s)
  real (4), dimension (3), intent(in) :: v
  type (Tspherical)                   :: s
  real (4), dimension (3) :: x

  x = normalize_vector(v)
  s.lo = atan2( x(2), x(1) )
  if ( x(3) >  1.0_SP ) x(3) = 1.0_SP
  if ( x(3) < -1.0_SP ) x(3) =-1.0_SP
  s.co = acos( x(3) )
end function uvector_to_spherical

!-------------------------------------------------------------------------------
end module math_operators
