program mim_launcher

  ! Program name: mim60
  !
  ! Necessary files : specify_file_names.f90, math_operators.f90,
  !                   fault_analyis.f90, mim_launcher.f90 (This file)
  !
  ! Purpose:    Launcher of the main processor of the Multiple Inverse Method.
  !
  ! Function:   A command-line argument is used to tell the name of the text 
  !             file that contains file names, k-values and other parameters 
  !             for the method. 
  !             This program is launched from the program minShell.exe, which
  !             has the GUI for specifying the parameters and the functions to
  !             verify the parameters. 
  !             
  ! Syntax:     mim70.exe project.txt
  !
  ! Example of project file:
  !             C:\aamyd\dev\mim\MI_mains\mim70\v1+v2+v3.fdt
  !             C:\aamyd\dev\mim\MI_mains\mim70\v123_k5r.mi4
  !             5
  !             .fdt
  !             5000
  !             FIXED
  !
  !
  ! Programmer: original version by A. Yamaji
  !             from ver. 3.2 to 4.0 by K. Sato
  !             from ver. 5.1 to 5.2 by M. Otsubo
  !             (Division of Earth & Planetary Sciences, Kyoto Univ.)
  !             from ver. 5.2 to 5.3 by M. Otsubo
  !             from ver. 5.2 to 5.5 by A. Yamaji
  !             from ver. 5.5 to 6.0 by A. Yamaji
  !             from ver. 6.0 to 6.1 by A. Yamaji
  !
  ! Version:     7.0  (January, 2011)  A. Yamaji
  !              6.1  (April, 4, 2010) A. Yamaji
  !              6.0  (March, 2010) A. Yamaji
  !              5.5  (March, 2010) A. Yamaji
  !              5.31 (March 22, 2005) M. Otsubo
  !              5.3  (March 22, 2005) M. Otsubo and K. Sato
  !                   change "exit" to "cycle" in subroutine MIM_kx
  !                   adjust output of loopCount, and k=3 to 8
  !              5.2  (February 28, 2005) M. Otsubo and A. Yamaji
  !                   k= 4 to 8, and cgp file name and misfit threshold are
  !                   collectively read from "pref.txt".
  !                   The function of ver5.1 is removed.
  !              5.1  (February 28, 2005) A. Yamaji
  !                   no vote if fault normal is near stress axis 
  !              5.0  (January 31, 2005)  A. Yamaji
  !                   no vote if heterogeneous
  !              4.0  (September,  2004)  K. Sato
  !              3.2                      A. Yamaji
  !
  ! Reference:  
  !             Otsubo, M. and Yamaji, A., 2006. Comput. Geosci., 32, 1221-1227. 
  !             Otsubo, M., Yamaji, A., Kubo, A., 2008. Tectonophys., 457, 150-160.
  !             Sato, K., Yamaji, A., 2006. J. Struct. Geol., 28, 972-979.
  !             Yamaji, A., 2000. J. Struct. Geol., 22, 441-452. 
  
!/////////////////////////// Decraration Section ///////////////////////////////

  use IFPORT
  use fault_analysis
  use specify_file_names, only : get_path_filename_extension
  use math_operators,     only : binomial_coef

  implicit none
  
  character(len=300)  :: DataFile
  character(len=300)  :: ResultFile
  character(len=300)  :: HomeDirectory
  character(len=300)  :: prjfile
  integer (1)         :: kValue
  integer (4)         :: Nrndm
  logical (1)         :: RndSeed = .FALSE.
  
  integer (1)         :: Narg
  character(256)      :: G_or_S
  logical             :: exists
  
  character(256)      :: name
  character(256)      :: ext
  character(len=4), parameter :: Default_ot_ext = '.mi4'

!/////////////////////////// Execution Section /////////////////////////////////

  call get_home_directory( HomeDirectory )
  call show_credit_title()
  
  Narg = get_n_cmd_args()
  if (Narg < 0) stop
  if (Narg == 0) then
    prjfile = 'project.txt'
  else
    call get_command_argument(1,prjfile)
    inquire(file=prjfile, exist=exists)
    if (.not.exists) then 
      write(*,*) '*** Error ***   project file not found'
      stop
    end if
  end if
  if ( .not. project_file_successfully_loaded() ) stop  
  ! echo back parameters
  call get_path_filename_extension ( DataFile,   name, ext )
  write(*,*) '  DataFile = ', trim(name)//trim(ext)
  call get_path_filename_extension ( ResultFile, name, ext )
  write(*,*) '  SaveFile = ', trim(name)//trim(ext)
  write(*,*) '  k        = ', kValue
  if (Nrndm>0) then 
    write(*,*) '  Nrandom  = ', Nrndm
    if (RndSeed) write(*,*) '  Initialize random number generator with date and time.'
  end if
  write(*,*)
  
  call MIM_main( HomeDirectory, DataFile, ResultFile, kValue, G_or_S, Nrndm, RndSeed )
  
contains !///////////////////// Subprograms ////////////////////////////////////
  
!-------------------------------------------------------------------------------
subroutine show_credit_title()
  print*
  print*, "  *************************************************************************"
  print*, "  *                                                                       *"
  print*, "  *   Multiple Inverse Method                                             *"
  print*, "  *                                                                       *"
  print*, "  *   Main Processor, Version 7.0                                         *"
  print*, "  *                                                                       *"
  print*, "  *   Produced by A. Yamaji, K. Sato and M. Otsubo                        *"
  print*, "  *   Division of Earth & Planetary Sciences, Kyoto University            *"
  print*, "  *                                                                       *"
  print*, "  *************************************************************************"
  print*
end subroutine show_credit_title

!-------------------------------------------------------------------------------
function check_cmd_args(datafile_ext) result(ok)
  character(len=4), intent(out) :: datafile_ext
  logical (1)         :: ok
  character(len =600) :: cmd_params
  integer (1)         :: cmd_n
  character(len=3)    :: drv
  character(len=300)  :: dir, fname_in, fname_ot
  character(len=4)    :: ext
  integer             :: i, len, status, dir_length
  logical (1)         :: exists

  ok = .FALSE.
  ! home directory and data file
  call get_command_argument(1, cmd_params, len, status);
  DataFile = cmd_params(1:len)
  dir_length = SPLITPATHQQ( DataFile, drv, dir, fname_in, datafile_ext )
  write(*,*)
  write(*,'(x,a)') 'Dir   = ' // trim(drv) // trim(dir)
  write(*,'(x,a)') 'Input = ' // trim(fname_in) // datafile_ext
  inquire( directory = (trim(drv) // trim(dir)), exist=exists )
  if (.not.exists) then
    write(*,*) '*** Error ***   directory not found'
    return
  end if
  inquire(file=DataFile, exist=exists)
  if (.not.exists) then
    write(*,*) '*** Error ***   data file not found'
    return
  end if
  
  ! outputfile
  call get_command_argument(2, cmd_params, len, status);
  ResultFile = cmd_params(1:len)
  dir_length = SPLITPATHQQ( ResultFile, drv, dir, fname_ot, ext )
  write(*,'(x,a)') 'Output= ' // trim(fname_ot) // ext
  if ((ext /= Default_ot_ext) .and. (ext /= '.MI4')) then
    write(*,*) '*** Error ***   invalid output file'
    return
  end if
  
  ! k-value
  call get_command_argument(3, cmd_params, len, status);
  read( cmd_params(1:len), '(i)', iostat=status ) kValue
  if ( (status > 0) .or. (kValue < 3) .or. (kValue > 8) ) then
    write(*,*) '*** Error ***   invalid k-value'
    return
  end if
  write(*,'(x,a)') 'k     = ' // cmd_params(1:len)
  
  ! number of random sampling
  
  if (Narg < 4) then
    Nrndm = 0
  else
    call get_command_argument(4, cmd_params, len, status)
    read( cmd_params(1:len), '(i)', iostat=status ) Nrndm
    if (status /= 0) then
      write(*,*) '*** Error ***   invalid M-value'
      return
    end if
  end if
  if (Nrndm==0) then
    write(*,*) 'NRS   = *'
  else 
    write(*,'(x,a,i<floor(log10(real(Nrndm)))+1>)') 'NRS   = ', Nrndm
  end if
  
  if (Narg >= 5) then
    call get_command_argument(5, cmd_params, len, status)
    RndSeed = (cmd_params(1:len)=='RND')
    if (RndSeed) then 
      write(*,*) 'Seed  = Random'
    else
      write(*,*) 'Seed  = Prescribed'
    end if
  end if
  
  ok = .TRUE.
end function check_cmd_args

!-------------------------------------------------------------------------------
function get_n_cmd_args() result(na)
  integer (1) :: na  
  character(len=1000) :: cmd_entire  
  integer             :: len, status
  
  call get_command(cmd_entire, len, status)  
  if (status /= 0) then
    write (*,*) '*** Error ***   invalid arguments', status 
    na = -1    
    return  
  end if  
  na = command_argument_count()
end function get_n_cmd_args

!-------------------------------------------------------------------------------
function read_mi_main_args result(ok)
  logical (1)                    :: ok
  character(len=16), parameter :: argfile = 'mi_main_args.txt'
  logical (1)                    :: exists
  integer (1)                    :: status
  integer                        :: dir_length
  character(len=3)               :: drv
  character(len=4)               :: ext
  character(len=400)             :: dir, fn1, fn2
  
  ok = .FALSE.
  write(*,*)
  
  inquire( file=argfile, exist=exists )
  if (.not. exists) then
    write(*,*) '*** Error ***   ' // argfile // ' not found'
    return
  end if
  
  open( unit=1, file=argfile )
  
  read(1,'(a)', iostat=status) dir
  inquire( directory=dir, exist=exists )
  if ((status /= 0) .or. (.not. exists)) then
    write(*,*) '*** Error ***   invalid home directory'
    return
  end if
  write(*,'(x,a)') 'Dir   = ' // trim(dir)
  
  read(1, '(a)', iostat=status) fn1
  DataFile = trim(dir) // '\' // trim(fn1)
  inquire( file=DataFile, exist=exists )
  if ((status /= 0) .or. (.not. exists)) then
    write(*,*) '*** Error ***   data file not found'
    return
  end if
  write(*,'(x,a)') 'Input = ' // trim(fn1)
  
  read(1, '(a)', iostat=status) fn2
  ResultFile = trim(dir) // '\' // trim(fn2)
  dir_length = SPLITPATHQQ( ResultFile, drv, fn1, fn2, ext )
  if ((ext /= '.mi4') .and. (ext /= 'MI4')) then
    write(*,*) '*** Error ***   output file has an invalid extension  : ' // ext
    return
  end if  
  write(*,'(x,a)') 'Output= ' // trim(fn2) // ext
  
  read(1, '(i)', iostat=status) kValue
  if ((status /=0) .or. (kValue < 3) .or. (kValue > 8)) then
    write(*,*) '*** Error ***   invalid k-value : ', kValue
    return
  end if
  write(*,'(x,a,i1)') 'k     = ', kValue
  
  read(1, '(i)', iostat=status) Nrndm
  if ((status /=0) .or. (Nrndm>Huge(Nrndm))) then
    write(*,*) '*** Error ***   invalid M-value : ', Nrndm
    return
  end if
  if (Nrndm==0) then
    write(*,*) 'M     = *'
  else 
    write(*,'(x,a,i<floor(log10(real(Nrndm)))+1>)') 'M     = ', Nrndm
  end if
  
  close(unit=1)
  
  ok = .TRUE.
end function read_mi_main_args

!-------------------------------------------------------------------------------
subroutine get_home_directory( dr )
  character(len=300), intent(out) :: dr ! directory
  integer (4)         :: er
  CHARACTER($MAXPATH) :: dir

  dir = FILE$CURDRIVE
  er  = GETDRIVEDIRQQ(dir)
  if (er.eq.0) then
    write (*,*) "*** Could'nt get home directory ***"
	write (*,*) "***     Execution terminated    ***"
	stop
  end if
  dr  = dir
end subroutine get_home_directory

!-------------------------------------------------------------------------------
function project_file_successfully_loaded() result(ok)
  logical :: ok
  integer :: ier
  character (len=1) :: randomseed
  ok = .TRUE.
  open( unit=1, file=prjfile, action='read' )
    
  read( unit=1, fmt=('(a)'), iostat=ier) DataFile
  if (ier/=0) then 
    ok = .FALSE.
    return
  end if  
  read( unit=1, fmt=('(a)'), iostat=ier) ResultFile
  if (ier/=0) then 
    ok = .FALSE.
    return
  end if  
  read( unit=1, fmt=('(i)'), iostat=ier) kValue
  if (ier/=0) then 
    ok = .FALSE.
    return
  end if  
  read( unit=1, fmt=('(a4)'), iostat=ier) G_or_S
   if (ier/=0) then 
    ok = .FALSE.
    return
  end if  
  read( unit=1, fmt=('(i)'), iostat=ier) Nrndm
   if (ier/=0) then 
    ok = .FALSE.
    return
  end if
  if (Nrndm==0) return 
  read( unit=1, fmt=('(a1)'), iostat=ier) randomseed
   if (ier/=0) then 
    ok = .FALSE.
    return
  end if
  RndSeed = .not. ((randomseed=='F') .or. (randomseed=='f'))
  
  close(unit=1)
end function project_file_successfully_loaded

!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
end program mim_launcher
